#!/bin/sh -xe

mkdir -p /usr/local/src
cd /usr/local/src

export PKG_CONFIG_PATH=/usr/local/lib/pkgconfig

git clone https://codeberg.org/DNS-OARC/dnswire.git
cd dnswire
git checkout v0.4.0
git submodule update --init
sh autogen.sh
./configure --disable-examples
make
make install
