#!/bin/sh -xe

mkdir -p /usr/local/src
cd /usr/local/src

wget -O tinyframe-0.1.1.tar.gz https://codeberg.org/DNS-OARC/tinyframe/archive/v0.1.1.tar.gz
tar zxf tinyframe-0.1.1.tar.gz
cd tinyframe
sh autogen.sh
./configure
make
make install
